package com.prowidesoftware.brilliance.spi;


import com.prowidesoftware.brilliance.util.ClassUtil;
import com.prowidesoftware.brilliance.util.PropertyUtil;
import com.prowidesoftware.brilliance.util.ReflectionUtil;
import com.prowidesoftware.brilliance.util.StringUtil;
import org.springframework.core.OrderComparator;
import org.springframework.util.Assert;

import java.io.IOException;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;

/**
 * spi加载类
 */
public class BeSpiLoader {
    public static final String SERVICES_RESOURCE_LOCATION = "META-INF/be-esb.services";
    public static final String COMMA_DELIMITED = ",";

    static final com.prowidesoftware.brilliance.spi.ConcurrentReferenceMap<ClassLoader, Map<String, List<String>>> cache = new ConcurrentReferenceMap<ClassLoader, Map<String, List<String>>>();


    private BeSpiLoader() {
    }

    public static <T> List<T> loadBeServices(Class<T> serviceType, ClassLoader classLoader) {
        return loadBeServices(serviceType, classLoader, null, null);
    }

    public static <T> List<T> loadBeServices(Class<T> serviceType, ClassLoader classLoader, Object[] args, Class[] argsTypes) {
        Assert.notNull(serviceType, "'serviceType' must not be null");
        ClassLoader classLoaderToUse = classLoader;
        if (classLoaderToUse == null) {
            classLoaderToUse = BeSpiLoader.class.getClassLoader();
        }
        List<String> serviceImplementationNames = loadServiceNames(serviceType, classLoaderToUse);
        List<T> result = new ArrayList<>(serviceImplementationNames.size());
        for (String serviceImplementationName : serviceImplementationNames) {
            result.add(instantiateService(serviceImplementationName, serviceType, classLoaderToUse, args, argsTypes));
        }
        OrderComparator.sort(result);
        return result;
    }

    public static List<String> loadServiceNames(Class<?> serviceType, ClassLoader classLoader) {
        ClassLoader classLoaderToUse = classLoader;
        if (classLoaderToUse == null) {
            classLoaderToUse = BeSpiLoader.class.getClassLoader();
        }
        String serviceTypeName = serviceType.getName();
        return loadBeServices(classLoaderToUse).getOrDefault(serviceTypeName, Collections.emptyList());
    }

    private static Map<String, List<String>> loadBeServices(ClassLoader classLoader) {
        Map<String, List<String>> result = cache.get(classLoader);
        if (result != null) {
            return result;
        }
        try {
            Enumeration<URL> urls = classLoader.getResources(SERVICES_RESOURCE_LOCATION);
            result = new HashMap<String, List<String>>();
            while (urls.hasMoreElements()) {
                URL url = urls.nextElement();
                Properties properties = PropertyUtil.loadProperties(url);
                for (Map.Entry<?, ?> entry : properties.entrySet()) {
                    String serviceTypeName = ((String) entry.getKey()).trim();
                    String[] serviceImplementationNames = ((String) entry.getValue()).split(COMMA_DELIMITED);
                    for (String serviceImplementationName : serviceImplementationNames) {
                        if(StringUtil.isEmpty(serviceImplementationName)){
                            continue;
                        }
                        result.computeIfAbsent(serviceTypeName, key -> new ArrayList<>())
                                .add(serviceImplementationName.trim());
                    }
                }
            }
            // 去重，并且转换为不能修改得list，按key为classLoader缓存到map
            result.replaceAll((serviceType, implementations) -> implementations.stream().distinct()
                    .collect(Collectors.collectingAndThen(Collectors.toList(), Collections::unmodifiableList)));
            cache.put(classLoader, result);
        } catch (IOException ex) {
            throw new IllegalArgumentException("Unable to load services from location [" +
                    SERVICES_RESOURCE_LOCATION + "]", ex);
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    private static <T> T instantiateService(String serviceImplementationName, Class<T> serviceType,
                                            ClassLoader classLoader, Object[] args, Class[] argsTypes) {
        try {
            Class<?> serviceImplementationClass = ClassUtil.forName(serviceImplementationName, classLoader);
            if (!serviceType.isAssignableFrom(serviceImplementationClass)) {
                throw new IllegalArgumentException(
                        "Class [" + serviceImplementationName + "] is not assignable to type [" + serviceType.getName() + "]");
            }
            if (args == null) {
                return (T) ReflectionUtil.accessibleConstructor(serviceImplementationClass).newInstance();
            } else {
                try{
                    return (T) ReflectionUtil.accessibleConstructor(serviceImplementationClass, argsTypes).newInstance(args);
                }catch (NoSuchMethodException e){
                    //尝试采用无参构造函数创建实例
                    return (T) ReflectionUtil.accessibleConstructor(serviceImplementationClass).newInstance();
                }
            }
        } catch (Throwable ex) {
            throw new IllegalArgumentException(
                    "Unable to instantiate class [" + serviceImplementationName + "] for type [" + serviceType.getName() + "]",
                    ex);
        }
    }

}
